library nui_log;

import 'dart:convert';
import 'dart:io';

import 'package:appcenter/appcenter.dart';
import 'package:appcenter_analytics/appcenter_analytics.dart';
import 'package:appcenter_crashes/appcenter_crashes.dart';
import 'package:flutter/cupertino.dart';
import 'package:nui_auth/nui_auth.dart';
import 'package:nui_core/code/nui_device_util.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_database/nui_database.dart';
import 'package:nui_error_handler/nui_error_handler.dart';
import 'package:nui_log/nui_log_database.dart';
import 'package:nui_log/nui_log_json.dart';
import 'package:nui_log/nui_log_models.dart';
import 'package:nui_web/nui_web.dart';
import 'nui_log_database.dart';

export 'package:nui_log/nui_log_database.dart';
export 'package:nui_log/nui_log_json.dart';
export 'package:nui_log/nui_log_models.dart';

typedef NUILogGetter = NUILogBuilder Function();

void runNUIAppWithLog(NUIApp app, {@required NUILogGetter logBuilder, @required ErrorHandlerGetter errorHandlerBuilder}) async{
  app.addNewTask(() => logBuilder().build());
  runNUIAppWithErrorHandler(app, errorHandlerBuilder, initListener: (){

  }, exceptionListener: (error, stacktrace) async{
    if(NUILog._instance != null) {
      NUILog._instance.saveLogFromStackTrace(error: error, stackTrace: stacktrace);
    }
  });
}

abstract class NUILogInterface{
  Future<NUILogCheckSettingResponse> checkLogSetting();
  Future<bool> saveLogs(List<NUILogContent> logs);
  Future<bool> saveLog(NUILogContent log);
  Future<NUILogContent> saveLogFromStackTrace({StackTrace stackTrace, dynamic error});
}

class NUILog implements NUILogInterface{
  static const _WEB_MODULE = "NUILogWebModule";
  static const _MODULE = "NUILog";
  static const _PREF_SETTING_KEY = "NUILogSettingKey";

  static NUILog _instance;

  String _baseUrl;
  String _appVersion;
  String _appId;
  String _appCenteriOSKey;
  String _appCenterAndroidKey;
  bool _appCenterInitialized = false;

  NUILog._builder(NUILogBuilder builder){
    _baseUrl = builder._baseUrl;
    _appVersion = builder._appVersion;
    _appId = builder._appId;
    _appCenterAndroidKey = builder._appCenterAndroidKey;
    _appCenteriOSKey = builder._appCenterIOSKey;

    if(Platform.isAndroid && !isNullOrEmpty(_appCenterAndroidKey)){
      logNUI(_MODULE, "Starting AppCenter for Android with key : $_appCenterAndroidKey");
      //Initialize AppCenter for Android
      AppCenter.start(_appCenterAndroidKey, [AppCenterAnalytics.id, AppCenterCrashes.id]);
      _appCenterInitialized = true;
    }
    if(Platform.isIOS && !isNullOrEmpty(_appCenteriOSKey)){
      logNUI(_MODULE, "Starting AppCenter for iOS with key : $_appCenterAndroidKey");
      //Initialize AppCenter for iOS
      AppCenter.start(_appCenteriOSKey, [AppCenterAnalytics.id, AppCenterCrashes.id]);
      _appCenterInitialized = true;
    }

    _instance = this;
  }

  static NUILog get(){
    if(_instance == null) throwNUI(_MODULE, "NUILog has not been built with builder");
    return _instance;
  }

  @override
  Future<NUILogContent> saveLogFromStackTrace({StackTrace stackTrace, dynamic error}) async{
    final deviceInfo = await NUIDeviceUtil.getDeviceInfo();

    var logContent = NUILogContent(
        id: randomUUID(),
        action: NUILogActionType.OTHER.code(),
        error: NUILogValueType.ERROR.code(),
        bundleId: _appId,
        hardwareModel: deviceInfo.model,
        osVersion: deviceInfo.version,
        happen: DateTime.now(),
        appVersion: _appVersion);

    //Extract information from exception
    if(stackTrace != null){
      logContent.module = NUINavigator.getCurrentScreenCode();
      logContent.subModule = error?.toString();
      logContent.description = stackTrace.toString();
    }

    //Get user id and session token from NUIAuth
    try{
      final userId = NUIAuth.get().getUserId();
      final sessionId = NUIAuth.get().getAccessToken()?.token;
      logContent.userId = userId;
      logContent.sessionId = sessionId;
    }catch (e){
      logNUI(_MODULE, "Failed to obtain session from NUIAuth");
    }

    logContent.logLevel = NUILogLevel.ERROR.value();

    saveLogs([logContent]);

    //Save exception to AppCenter
    if(stackTrace != null && _appCenterInitialized == true){
      final deviceInfo = await getDeviceInfo();
      AppCenterAnalytics.trackEvent((NUINavigator.getCurrentScreenCode().isNullOrEmpty() ? "Exception" : ("Exception at " + NUINavigator.getCurrentScreenCode())) + " v$_appVersion" + " - " + deviceInfo.model, logContent.toLogMap()); // with custom properties
    }

    return logContent;
  }

  @override
  Future<NUILogCheckSettingResponse> checkLogSetting() async{
    final result = await NUIWeb.get(module: _WEB_MODULE).doPOST<NUILogCheckSettingResponse>(url: "logging/setting", entity: NUILogCheckSettingResponseEntity());
    if(result.success == true){
      logNUI(_MODULE, "Check log setting successful");
      NUISharedPref.get().save(_PREF_SETTING_KEY, result.body);
      return result.body;
    }
    else{
      logNUI(_MODULE, "Check log failed with error : ${result.message}");
      return null;
    }
  }

  Future<NUILogSetting> _getSettingByLevel(int logLevel) async{
    final setting = await NUISharedPref.get().getObject(_PREF_SETTING_KEY, NUILogCheckSettingResponseEntity());
    if(setting == null) return null;

    NUILogSetting logSetting;
    for(NUILogSetting setting in setting.data){
      if(setting.logLevel == logLevel) logSetting = setting;
      break;
    }

    return logSetting;
  }

  @override
  Future<bool> saveLog(NUILogContent log) async{
    return await saveLogs([log]);
  }

  @override
  Future<bool> saveLogs(List<NUILogContent> logs) async{
    final filteredLogs = List<NUILogContent>();
    for(NUILogContent log in logs){
      if(log.logLevel == null){
        filteredLogs.add(log);
        continue;
      }

      final logSetting = await _getSettingByLevel(log.logLevel);
      if(logSetting == null){
        filteredLogs.add(log);
      }
      else{
        if(logSetting.enable == true){
          filteredLogs.add(log);
        }
      }
    }

    if(isNullOrEmpty(filteredLogs)) {
      logNUI(_MODULE, "No logs to be saved according to the settings");
      return false;
    }

    if(_appCenterInitialized == true) {
      logNUI(_MODULE, "Saving ${logs.length} logs to AppCenter");
      filteredLogs.forEach((element) async {
        final deviceInfo = await getDeviceInfo();
        element.logLevel = element.logLevel ?? NUILogLevel.DEBUG.value();
        final title = (NUINavigator.getCurrentScreenCode() ?? "Log") + " v$_appVersion" + " - " + deviceInfo.model;
        logNUI(_MODULE, "Saving log : $title, content : ${jsonEncode(element)}");
        AppCenterAnalytics.trackEvent(title, element.toLogMap()); // with custom properties
      });
    }

    final request = NUILogSaveLogRequest(list: filteredLogs);
    final result = await NUIWeb.get(module: _WEB_MODULE).doPOST(url: "logging/save", bodyEnt: request);
    if(result.success == true){
      final innerSuccess = result.responseBody["success"];
      if(innerSuccess == true) {
        logNUI(_MODULE, "Save logs successful");
        return true;
      }
      else{
        return false;
      }
    }
    else{
      logNUI(_MODULE, "Failed to save logs with error : ${result.message}");
      return false;
    }
  }
}

class NUILogBuilder{
  String _baseUrl;
  String _appVersion;
  String _appId;
  String _appCenterIOSKey;
  String _appCenterAndroidKey;

  NUILogBuilder baseUrl(String baseUrl){
    _baseUrl = baseUrl;
    return this;
  }

  NUILogBuilder appCenteriOSKey(String key){
    _appCenterIOSKey = key;
    return this;
  }

  NUILogBuilder appCenterAndroidKey(String key){
    _appCenterAndroidKey = key;
    return this;
  }

  NUILogBuilder appVersion(String appVersion){
    _appVersion = appVersion;
    return this;
  }

  NUILogBuilder appId(String appId){
    _appId = appId;
    return this;
  }

  Future<NUILog> build() async{
    final db = NUIDatabaseBuilder()
        .model(NUILogDatabase())
        .build();
    final pref = NUISharedPrefBuilder()
        .build();
    final web = NUIWebBuilder()
      .module(NUILog._WEB_MODULE)
      .baseURL(_baseUrl)
      .build();

    await db;
    await pref;
    return NUILog._builder(this);
  }
}
